// Copyright (c) .NET Foundation. All rights reserved.
// Licensed under the Apache License, Version 2.0. See License.txt in the project root for license information.

using System.IO;
using JetBrains.Annotations;
using Microsoft.Data.Entity.Metadata;
using Microsoft.Data.Entity.Relational;
using Microsoft.Data.Entity.Relational.Migrations.Infrastructure;
using Microsoft.Data.Entity.Relational.Migrations.Sql;
using Microsoft.Data.Entity.Utilities;


namespace Microsoft.Data.Entity.Sqlite
{
    public class SqliteDataStoreCreator : RelationalDataStoreCreator
    {
        private const int SQLITE_CANTOPEN = 14;

        private readonly IRelationalConnection _connection;
        private readonly IModelDiffer _modelDiffer;
        private readonly IMigrationSqlGenerator _migrationSqlGenerator;
        private readonly ISqlStatementExecutor _executor;

        public SqliteDataStoreCreator(
            [NotNull] IRelationalConnection connection,
            [NotNull] IModelDiffer modelDiffer,
            [NotNull] IMigrationSqlGenerator migrationSqlGenerator,
            [NotNull] ISqlStatementExecutor sqlStatementExecutor)
        {
            Check.NotNull(connection, nameof(connection));
            Check.NotNull(modelDiffer, nameof(modelDiffer));
            Check.NotNull(migrationSqlGenerator, nameof(migrationSqlGenerator));
            Check.NotNull(sqlStatementExecutor, nameof(sqlStatementExecutor));

            _connection = connection;
            _modelDiffer = modelDiffer;
            _migrationSqlGenerator = migrationSqlGenerator;
            _executor = sqlStatementExecutor;
        }

        public override void Create()
        {
            _connection.Open();
            _connection.Close();
        }

        public override void CreateTables(IModel model)
        {
            Check.NotNull(model, nameof(model));

            var operations = _modelDiffer.GetDifferences(null, model);
            var statements = _migrationSqlGenerator.Generate(operations, model);
            _executor.ExecuteNonQuery(_connection, null, statements);
        }


        public override bool Exists() => true;

        public override bool HasTables()
        {
            var count = (long)_executor.ExecuteScalar(
                _connection,
                null,
                "SELECT COUNT(*) FROM \"sqlite_master\" WHERE \"type\" = 'table' AND \"rootpage\" IS NOT NULL;");

            return count != 0;
        }

        public override void Delete()
        {
            string path = null;

            _connection.Open();
            try
            {
                path = _connection.DbConnection.DataSource;
                _connection.Close();
            }
            catch
            {
            }

            if (path != null)
            {
                File.Delete(path);
            }
        }
    }
}
