// Copyright (c) .NET Foundation. All rights reserved.
// Licensed under the Apache License, Version 2.0. See License.txt in the project root for license information.

using System;
using System.Globalization;
using System.Threading;
using Xunit;

namespace Microsoft.Data.Entity.Commands.Utilities
{
    public class CSharpHelperTest : IDisposable
    {
        private readonly CultureInfo _backupCultureInfo;

        public CSharpHelperTest()
        {
            _backupCultureInfo = Thread.CurrentThread.CurrentCulture;
            // Ensure that there is a culture used which differs from the intended formatting and also from typical CI machines.
            // This way, formatting dependent on the local system's settings will be caught by the tests.
            Thread.CurrentThread.CurrentCulture = CultureInfo.GetCultureInfo("de-DE");
        }

        void IDisposable.Dispose()
        {
            Thread.CurrentThread.CurrentCulture = _backupCultureInfo;
        }
        [Theory]
        [InlineData(
            "single-line string with \"",
            "\"single-line string with \\\"\"")]
        [InlineData(
            true,
            "true")]
        [InlineData(
            false,
            "false")]
        [InlineData(
            (byte)42,
            "(byte)42")]
        [InlineData(
            'A',
            "'A'")]
        [InlineData(
            '\'',
            @"'\''")]
        [InlineData(
            4.2,
            "4.2")]
        [InlineData(
             -1.7976931348623157E+308, // Double MinValue
            "-1.7976931348623157E+308")]
        [InlineData(
             1.7976931348623157E+308, // Double MaxValue
            "1.7976931348623157E+308")]
        [InlineData(
            4.2f,
            "4.2f")]
        [InlineData(
             -3.402823E+38f, // Single MinValue
            "-3.402823E+38f")]
        [InlineData(
             3.402823E+38f, // Single MaxValue
            "3.402823E+38f")]
        [InlineData(
            42,
            "42")]
        [InlineData(
            42L,
            "42L")]
        [InlineData(
             9000000000000000000L, // Ensure not printed as exponent
            "9000000000000000000L")]
        [InlineData(
            (sbyte)42,
            "(sbyte)42")]
        [InlineData(
            (short)42,
            "(short)42")]
        [InlineData(
            42u,
            "42u")]
        [InlineData(
            42ul,
            "42ul")]
        [InlineData(
             18000000000000000000ul, // Ensure not printed as exponent
            "18000000000000000000ul")]
        [InlineData(
            (ushort)42,
            "(ushort)42")]
        public void Literal_works(object value, string expected)
        {
            var literal = new CSharpHelper().UnknownLiteral(value);
            Assert.Equal(expected, literal);
        }

        [Fact]
        public void Literal_works_when_empty_ByteArray() =>
            Literal_works(
                new byte[0],
                "new byte[] {  }");

        [Fact]
        public void Literal_works_when_single_ByteArray() =>
            Literal_works(
                new byte[] { 1 },
                "new byte[] { 1 }");

        [Fact]
        public void Literal_works_when_many_ByteArray() =>
            Literal_works(
                new byte[] { 1, 2 },
                "new byte[] { 1, 2 }");

        [Fact]
        public void Literal_works_when_multiline_string() =>
            Literal_works(
                "multi-line" + Environment.NewLine + "string with \"",
                "@\"multi-line" + Environment.NewLine + "string with \"\"\"");

        [Fact]
        public void Literal_works_when_DateTime() =>
            Literal_works(
                new DateTime(2015, 3, 15, 20, 45, 17, 300, DateTimeKind.Local),
                "new DateTime(2015, 3, 15, 20, 45, 17, 300, DateTimeKind.Local)");

        [Fact]
        public void Literal_works_when_DateTimeOffset() =>
            Literal_works(
                new DateTimeOffset(new DateTime(2015, 3, 15, 19, 43, 47, 500), new TimeSpan(-7, 0, 0)),
                "new DateTimeOffset(new DateTime(2015, 3, 15, 19, 43, 47, 500, DateTimeKind.Unspecified), new TimeSpan(0, -7, 0, 0, 0))");

        [Fact]
        public void Literal_works_when_decimal() =>
            Literal_works(
                4.2m,
                "4.2m");

        [Fact]
        public void Literal_works_when_decimal_max_value() =>
            Literal_works(
                 79228162514264337593543950335m, // Decimal MaxValue
                "79228162514264337593543950335m");

        [Fact]
        public void Literal_works_when_decimal_min_value() =>
            Literal_works(
                 -79228162514264337593543950335m, // Decimal MinValue
                "-79228162514264337593543950335m");

        [Fact]
        public void Literal_works_when_Guid() =>
            Literal_works(
                new Guid("fad4f3c3-9501-4b3a-af99-afeb496f7664"),
                "new Guid(\"fad4f3c3-9501-4b3a-af99-afeb496f7664\")");

        [Fact]
        public void Literal_works_when_TimeSpan() =>
            Literal_works(
                new TimeSpan(17, 21, 42, 37, 250),
                "new TimeSpan(17, 21, 42, 37, 250)");

        [Fact]
        public void Literal_works_when_NullableInt() =>
            Literal_works(
                (int?)42,
                "42");

        [Fact]
        public void Literal_works_when_single_StringArray()
        {
            var literal = new CSharpHelper().Literal(new[] { "A" });
            Assert.Equal("\"A\"", literal);
        }

        [Fact]
        public void Literal_works_when_many_StringArray()
        {
            var literal = new CSharpHelper().Literal(new[] { "A", "B" });
            Assert.Equal("new[] { \"A\", \"B\" }", literal);
        }

        [Fact]
        public void UnknownLiteral_throws_when_unknown()
        {
            var ex = Assert.Throws<InvalidOperationException>(
                () => new CSharpHelper().UnknownLiteral(new object()));
            Assert.Equal(Strings.UnknownLiteral(typeof(object)), ex.Message);
        }

        [Fact]
        public void Reference_works_when_nested()
        {
            Assert.Equal(
                "CSharpHelperTest.Nested",
                new CSharpHelper().Reference(typeof(Nested)));
        }

        [Fact]
        public void Reference_works_when_nested_more()
        {
            Assert.Equal(
                "CSharpHelperTest.Nested.DoubleNested",
                new CSharpHelper().Reference(typeof(Nested.DoubleNested)));
        }

        private class Nested
        {
            public class DoubleNested
            {
            }
        }
    }
}