// Copyright (c) .NET Foundation. All rights reserved.
// Licensed under the Apache License, Version 2.0. See License.txt in the project root for license information.

using System;
using System.Linq;
using System.Threading.Tasks;
using Microsoft.Data.Entity.ChangeTracking.Internal;
using Microsoft.Data.Entity.Metadata;
using Microsoft.Data.Entity.Metadata.ModelConventions;
using Microsoft.Data.Entity.Tests;
using Microsoft.Framework.DependencyInjection;
using Microsoft.Framework.Logging;
using Moq;
using Xunit;

namespace Microsoft.Data.Entity.InMemory.Tests
{
    public class InMemoryDataStoreTest
    {
        [Fact]
        public void Uses_persistent_database_by_default()
        {
            var serviceProvider = InMemoryTestHelpers.Instance.CreateServiceProvider();

            var store1 = InMemoryTestHelpers.Instance.CreateContextServices(serviceProvider, CreateModel()).GetRequiredService<IInMemoryDataStore>();
            var store2 = InMemoryTestHelpers.Instance.CreateContextServices(serviceProvider, CreateModel()).GetRequiredService<IInMemoryDataStore>();

            Assert.Same(store1.Database, store2.Database);
        }

        [Fact]
        public void Uses_persistent_database_if_configured_as_persistent()
        {
            var serviceProvider = InMemoryTestHelpers.Instance.CreateServiceProvider();

            Assert.Same(
                CreateStore(serviceProvider, persist: true).Database,
                CreateStore(serviceProvider, persist: true).Database);
        }

        [Fact]
        public void Uses_transient_database_if_not_configured_as_persistent()
        {
            var serviceProvider = InMemoryTestHelpers.Instance.CreateServiceProvider();

            Assert.NotSame(
                CreateStore(serviceProvider, persist: false).Database,
                CreateStore(serviceProvider, persist: false).Database);
        }

        [Fact]
        public void EnsureDatabaseCreated_returns_true_for_first_use_of_persistent_database_and_false_thereafter()
        {
            var serviceProvider = InMemoryTestHelpers.Instance.CreateServiceProvider();
            var model = CreateModel();
            var store = CreateStore(serviceProvider, persist: true);

            Assert.True(store.EnsureDatabaseCreated(model));
            Assert.False(store.EnsureDatabaseCreated(model));
            Assert.False(store.EnsureDatabaseCreated(model));

            store = CreateStore(serviceProvider, persist: true);

            Assert.False(store.EnsureDatabaseCreated(model));
        }

        [Fact]
        public void EnsureDatabaseCreated_returns_true_for_first_use_of_non_persistent_database_and_false_thereafter()
        {
            var serviceProvider = InMemoryTestHelpers.Instance.CreateServiceProvider();
            var model = CreateModel();
            var store = CreateStore(serviceProvider, persist: false);

            Assert.True(store.EnsureDatabaseCreated(model));
            Assert.False(store.EnsureDatabaseCreated(model));
            Assert.False(store.EnsureDatabaseCreated(model));

            store = CreateStore(serviceProvider, persist: false);

            Assert.True(store.EnsureDatabaseCreated(model));
            Assert.False(store.EnsureDatabaseCreated(model));
            Assert.False(store.EnsureDatabaseCreated(model));
        }

        private static IInMemoryDataStore CreateStore(IServiceProvider serviceProvider, bool persist)
        {
            var optionsBuilder = new EntityOptionsBuilder();
            optionsBuilder.UseInMemoryStore(persist: persist);

            return InMemoryTestHelpers.Instance.CreateContextServices(serviceProvider, optionsBuilder.Options).GetRequiredService<IInMemoryDataStore>();
        }

        [Fact]
        public async Task Save_changes_adds_new_objects_to_store()
        {
            var serviceProvider = InMemoryTestHelpers.Instance.CreateContextServices(CreateModel());
            var customer = new Customer { Id = 42, Name = "Unikorn" };
            var entityEntry = serviceProvider.GetRequiredService<IStateManager>().GetOrCreateEntry(customer);
            entityEntry.SetEntityState(EntityState.Added);

            var inMemoryDataStore = serviceProvider.GetRequiredService<IInMemoryDataStore>();

            await inMemoryDataStore.SaveChangesAsync(new[] { entityEntry });

            Assert.Equal(1, inMemoryDataStore.Database.SelectMany(t => t).Count());
            Assert.Equal(new object[] { 42, "Unikorn" }, inMemoryDataStore.Database.Single().Single());
        }

        [Fact]
        public async Task Save_changes_updates_changed_objects_in_store()
        {
            var serviceProvider = InMemoryTestHelpers.Instance.CreateContextServices(CreateModel());

            var customer = new Customer { Id = 42, Name = "Unikorn" };
            var entityEntry = serviceProvider.GetRequiredService<IStateManager>().GetOrCreateEntry(customer);
            entityEntry.SetEntityState(EntityState.Added);

            var inMemoryDataStore = serviceProvider.GetRequiredService<IInMemoryDataStore>();

            await inMemoryDataStore.SaveChangesAsync(new[] { entityEntry });

            customer.Name = "Unikorn, The Return";
            entityEntry.SetEntityState(EntityState.Modified);

            await inMemoryDataStore.SaveChangesAsync(new[] { entityEntry });

            Assert.Equal(1, inMemoryDataStore.Database.SelectMany(t => t).Count());
            Assert.Equal(new object[] { 42, "Unikorn, The Return" }, inMemoryDataStore.Database.Single().Single());
        }

        [Fact]
        public async Task Save_changes_removes_deleted_objects_from_store()
        {
            var serviceProvider = InMemoryTestHelpers.Instance.CreateContextServices(CreateModel());

            var customer = new Customer { Id = 42, Name = "Unikorn" };
            var entityEntry = serviceProvider.GetRequiredService<IStateManager>().GetOrCreateEntry(customer);
            entityEntry.SetEntityState(EntityState.Added);

            var inMemoryDataStore = serviceProvider.GetRequiredService<IInMemoryDataStore>();

            await inMemoryDataStore.SaveChangesAsync(new[] { entityEntry });

            // Because the data store is being used directly the entity state must be manually changed after saving.
            entityEntry.SetEntityState(EntityState.Unchanged);

            customer.Name = "Unikorn, The Return";
            entityEntry.SetEntityState(EntityState.Deleted);

            await inMemoryDataStore.SaveChangesAsync(new[] { entityEntry });

            Assert.Equal(0, inMemoryDataStore.Database.SelectMany(t => t).Count());
        }

        [Fact]
        public async Task Should_log_writes()
        {
            var mockLogger = new Mock<ILogger>();
            mockLogger.Setup(l => l.IsEnabled(LogLevel.Information)).Returns(true);

            var mockFactory = new Mock<ILoggerFactory>();
            mockFactory.Setup(m => m.CreateLogger(It.IsAny<string>())).Returns(mockLogger.Object);

            var serviceCollection = new ServiceCollection();
            serviceCollection.AddInstance(mockFactory.Object);

            var scopedServices = InMemoryTestHelpers.Instance.CreateContextServices(serviceCollection, CreateModel());

            var customer = new Customer { Id = 42, Name = "Unikorn" };
            var entityEntry = scopedServices.GetRequiredService<IStateManager>().GetOrCreateEntry(customer);
            entityEntry.SetEntityState(EntityState.Added);

            var inMemoryDataStore = scopedServices.GetRequiredService<IInMemoryDataStore>();

            await inMemoryDataStore.SaveChangesAsync(new[] { entityEntry });

            mockLogger.Verify(
                l => l.Log(
                    LogLevel.Information,
                    0,
                    1,
                    null,
                    It.IsAny<Func<object, Exception, string>>()),
                Times.Once);
        }

        private static IModel CreateModel()
        {
            var modelBuilder = new ModelBuilder(new ConventionSet());

            modelBuilder.Entity<Customer>(b =>
                {
                    b.Key(c => c.Id);
                    b.Property(c => c.Name);
                });

            return modelBuilder.Model;
        }

        private class Customer
        {
            public int Id { get; set; }
            public string Name { get; set; }
        }
    }
}
